// Penn Parallel Primitives Library
// Author: Prof. Milo Martin
// University of Pennsylvania
// Spring 2010

#include <pthread.h>
#include <stdio.h>
#include "ppp.h"
#include "machine.h"
#include "random.h"

namespace ppp {

  namespace internal {
    static bool initialized = false;
    __thread int s_thread_id = 0;
    __thread int s_thread_count = 0;

    struct thread_info_t {
      pthread_t id;
      int thread_id;
      int thread_count;
      void (*thread_func) (void);
    };

  }
  void reseed_random_per_thread();
  uint32_t threadsafe_random();

  using namespace internal;
  
  // Helper function to take care of the void* conversion
  static void* init_start(void *ptr) 
  {
    thread_info_t* info_ptr = (thread_info_t *) ptr;    // cast from void*
    s_thread_id = info_ptr->thread_id;
    s_thread_count = info_ptr->thread_count;
    reseed_random_per_thread(); // Give each thread its own set of random numbers
    (info_ptr->thread_func)();  // Call through function pointer
    return NULL;  // terminates thread
  }

  void init(void (*func) (void), int num_threads)
  {
    // Only init threads once
    if (initialized) {
      printf("Error: init() called more than once\n");
      exit(1);
    }
    initialized = true;

    if (num_threads < 1) {
      printf("Error: number of threads passed to init() must be > 0\n");
      exit(1);
    }

    if (num_threads > MAX_THREADS) {
      printf("Error: number of threads (%d) exceeds MAX_THREADS (%d)\n", num_threads, MAX_THREADS);
      exit(1);
    }

    
    // Allocate and initialize the thread info array
    thread_info_t* info_array = new thread_info_t[num_threads];
    for (int i=0; i<num_threads; i++) {
      info_array[i].thread_id = i;
      info_array[i].thread_count = num_threads;
      info_array[i].thread_func = func;
    }
    
    // Create threads 1 through N-1
    for (int i=1; i<num_threads; i++) {
      int error = pthread_create(&(info_array[i].id), NULL, init_start, &info_array[i]);
      if (error) { 
        printf("Error: Thread creation error\n");
        exit(1);
      }
    }
    
    // "Create" thread zero
    init_start(&info_array[0]);
    
    // Wait for all threads to terminate
    for (int i=1; i<num_threads; i++) {
      pthread_join(info_array[i].id, NULL);
    }
    
    delete [] info_array;
  }

}


